package net.sergeych.myonly.cloud

import java.text.Normalizer

@Suppress("unused")
object UNS {

    fun reduce(source: String): String {
        // step 1: remove space and punctuation
        var name = source.lowercase().trim().replace(rePunctuation,"_")
        // step 2: NFKD
        name = Normalizer.normalize(name, Normalizer.Form.NFKD)
        // step 3: XLAT1: removing composing characters and ligatures
        name = name.map { xlat1.getOrDefault(it, it.toString()) }.joinToString("")
        // step 4: reduce to glyph archetype
        return name.map { ch ->
            xlat2.getOrElse(ch) {
                throw IllegalArgumentException("illegal character `$ch` in $name")
            }
        }.joinToString("")
    }

    val xlat1 by lazy { readXlat(DEFAULT_XLAT1, "") }

    val xlat2 by lazy {
        val result = readXlat(DEFAULT_XLAT2, null)
        readXlat(DEFAULT_XLAT2_FINALIZER, null).forEach { (finalKey, finalValue) ->
            // finalizer algorithm: if it overrides result's value, alter it
            // not effective at build time, but more effective when processing strings
            // update all existing values according to final update table
            val affectedKeys = result.filterValues { it[0] == finalKey }.keys
            for (k in affectedKeys) result[k] = finalValue
            result[finalKey] = finalValue
        }
        result
    }

    private fun decode(characterDefinition: String): Pair<Int, Char> {
        val ch = characterDefinition.trim()
        return if (ch.startsWith("U+")) {
            val code: Int = ch.substring(2).toInt(16)
            Pair(code, code.toChar())
        } else
            Pair(ch[0].code, ch[0])
    }

    private val reSpaceDelimiter = Regex("""\s+""")
    private val reRange = Regex("""^(.+):(.+)$""")
    private val reNotUchar = Regex("^(?!U\\+)")
    private val rePunctuation = Regex("""([-_=+()*&^%#@±§~`<,>/?'";:{}\[\]]|\s)+""")

    /**
     * @param missing if null, where applicable replace to self
     */
    private fun readXlat(xlat: String, missing: String?): MutableMap<Char, String> {
        val all = HashMap<Char, String>()
        for (l in xlat.lines()) {
            val line = l.split('#', limit = 2)[0].trim()
            try {
                if (line.isNotBlank()) {
                    val parts = line.split(reSpaceDelimiter)
                    val left = parts[0]
                    val right = parts.getOrNull(1)
                    reRange.find(left)?.let {
                        // from:to range
                        val start = decode(it.groupValues[1]).first
                        val stop = decode(it.groupValues[2]).first
                        for (code in start..stop) {
                            val ch = code.toChar()
                            all[ch] = right ?: (missing ?: ch.toString())
                        }
                    } ?: reNotUchar.find(left)?.let {
                        // set of chars
                        for (ch in left) {
                            all[ch] = right ?: (missing ?: ch.toString())
                        }
                    } ?: run {
                        // single U+code
                        all[decode(left).second] = right ?: run {
                            if (missing == null)
                                left
                            else
                                throw IllegalStateException("shouldn't be empty")
                        }
                    }
                }
            } catch (e: Exception) {
                println("error in line: $line")
                throw e
            }
        }
        return all
    }

    val DEFAULT_XLAT1 = """
        # Combining diacritical marks, see https://en.wikipedia.org/wiki/Combining_Diacritical_Marks

        U+0300:U+033C
        U+033D          x
        U+033E:U+0362
        U+0363          a
        U+0364          e
        U+0365          i
        U+0366          o
        U+0367          u
        U+0368          c
        U+0369          d
        U+036A          h
        U+036B          m
        U+036C          r
        U+036D          t
        U+036E          v
        U+036F          x

        #
        # replacing ligatures ------------------------
        #

        ꜳ            aa
        æ            ae
        ꜵ           ao
        ꜷ           au
        ꜹ           av
        ꜻ           av
        ꜽ           ay
        U+1F670     et
        ﬀ           ff
        ﬃ          ffi
        ﬄ          ffl
        ﬁ           phi
        ﬂ           fl
        œ           oe
        ꝏ          oo
        ß           ss
        ﬆ           st
        ﬅ           st
        ꜩ           tz
        ᵫ           ue
        ꝡ           vy

        """.trimIndent()

    val DEFAULT_XLAT2 = """ 
        #
        # Glyph archetypes
        #

        # Cyrillic letters

        а a
        б b

        в b  # May be similar in uppercase

        г 2  # In some fonts. RFC!
        д
        е e
        ж x
        з 3
        и u
        к k
        л n
        м m
        н h
        о o
        п n
        р p
        с c
        т t
        у y
        ф
        х x
        ц u
        ч 4
        ш
        щ ш
        ьъ b
        ы bi
        э 3
        ю io
        я 9

        # Other national languages for characters that will not be
        # normalized with NFKD to latin set. Do not put here any
        # characters with diacritic modifications removed by NFKD normalization.
        """.trimIndent()


    val DEFAULT_XLAT2_FINALIZER = """
        #
        # Final similarity corrections table
        #
        # English correcting similar-looking glyphs. This section MUST be the last
        # AND should altrer XLAT 2 table the following way:
        #
        # foreach ch in final
        # - if xlat2[*] = 

        # first init all with self, for simplicity:
        a:z
        0:9

        # Now replace similar glyphs to archetypes:

        il1|!  1
        o0øº   0
        u      v
        w      vv
        ${'$'}s             5
        b      6

        # Punctuation placeholder:
        _

        # Cluster delimiter
        .

    """.trimIndent()

    val excludedNames by lazy {
        listOf(
            "sergeych",
            "sergeych2",
            "real_sergeych",
            "real sergeych",
            "sergey chernov",
            "chernov sergey",
            "sergey s. chernov",
            "universa",
            "uni",
            "myo",
            "myonly",
            "myonly.cloud",
            "myonlycloud",
            "myonly.cloud",
            "myonly.cloud",
            "myonly.cloud"
        )
    }

    @Suppress("unused")
    val excludedReducedNames by lazy { excludedNames.map{ reduce(it) }.toSet() }

}